from distutils.core import setup

setup(name='Candela',
      version='0.2.2',
      description="Python shell creation library",
      author="Emmett Butler",
      author_email="emmett.butler321@gmail.com",
      url="https://github.com/emmett9001/candela",
      packages=['candela'],
      )
